/* This program solves the 0-1 Knapsack problem using Branch and Bound
 * For information on Branch and Bound and the Knapsack Problem
 * you can use the book of Horowitz and Sahni,
 * "Fundamentals of Computers Algorithms" (Pitman)
 * For more detailed information about the Knapsack problem, see
 * "Knapsack Problems. Algorithms and Computer Implementations." by
 * Silvano Martello and Paolo Toth. John Wiley and Sons.
 * This approach focuses in the aspects of general
 * parallel branch and bound and does not in the concrete
 * features of knapsack Problems.
 * Author: Casiano Rodriguez Leon. Email: casiano@ull.es
*/

#include <fork.h>
#include <io.h>
#include <stdlib.h>
#include <assert.h>   

#define MAXNUMBEROFOBJECTS 64 
#define MAXHEAPSIZE 1024 
#define STRIDE 1
#define  INFINITY 8192
#define HEAPFULL 0 
/* (((MAXHEAPSIZE-(tail-head))% MAXHEAPSIZE) >= MAXHEAPSIZE -1) */
/* INFINITY has to be set to a value greater 
   than the capacity M of the Knapsack */
typedef int number;
typedef int boolean;

typedef struct {
	int k, C, P;
} problemt;

#define ns "%d\n"
/* knapsack support variables */
sh number N, M;
sh number w[MAXNUMBEROFOBJECTS], p[MAXNUMBEROFOBJECTS];
/*heap maintenance variables */
number bestSol = -INFINITY;
sh problemt problems[MAXHEAPSIZE];
number nP; /* number of problems in the heap */
sh number head = 0, tail = 0;
/* variables for parallel cmputing */
sh int NUMPROCESSORS;

sync void readKnap(void) {
	FILE *in;
  number i;

	farm {
		if ((in = fopen("knap.dat","r+")) == NULL) {
		 fprintf(stderr, "Cannot open input file knap.dat.\n");
		 fprintf(stderr,"Input file format is:\nNumber of objects\nCapacity\nweight 0\nprofit 0\n...\n");
		 exit(1);
		}
		fscanf(in,ns,&N);
		fscanf(in,ns,&M);
		printf("N = "); printf(ns,N);
		printf("M = "); printf(ns,M);
		assert(INFINITY > M);
		for(i=0;i<N;i++) {
			fscanf(in,ns,&w[i]);
			fscanf(in,ns,&p[i]);
			if (i%STRIDE == 0) {
				printf("w = "); printf(ns,w[i]);
				printf("p = "); printf(ns,p[i]);
		 }
		}
		w[N] = INFINITY; p[N] = 0; /* sentinel object */
		fclose(in);
	}
}

sync void lowerUpper(problemt Prob, number *L, number *U) {
  /* naive approach */
  number i, weig = 0, prof = Prob.P;

  if (Prob.C < 0) {
	 *L = -INFINITY; *U = -INFINITY;
  }
  else {
	 for(i=Prob.k; weig <= Prob.C; i++) {
		weig += w[i]; prof += p[i];
	 }
	 /* invariant:  weig > prob.C */
	 i--;
	 weig -= w[i];
	 prof -= p[i];
	 *L = prof;  /* see Martello and Toth pag 16. Th 2.1 */
	 *U = prof+(p[i]*(Prob.C-weig))/w[i];
  }
}

sync void setProblem(problemt *Prob, number k, number C, number P) {

  Prob->k = k;
  Prob->C = C;
  Prob->P = P;
}

sync number numProblems(void) { /* number of problems in the queue */
  return (((tail-head)+MAXHEAPSIZE)%MAXHEAPSIZE);
}

sync void getProblem(problemt *Prob) {

	setProblem(Prob, 
			problems[head+$].k, problems[head+$].C, problems[head+$].P);

	NUMPROCESSORS = 0;
	$ = mpadd(&NUMPROCESSORS,1);
  if ($ == 0) {
		head = (head+NUMPROCESSORS)% MAXHEAPSIZE;
		farm printf("\n getting %d problems head = %d, tail = %d",
								NUMPROCESSORS,head,tail);
	}
}

sync void insertProblem(problemt Prob) {
  number next;

	NUMPROCESSORS = 0;
	$ = mpadd(&NUMPROCESSORS,1);
  next = (tail+NUMPROCESSORS)%MAXHEAPSIZE;

  if (HEAPFULL) farm { 
	 printf("\n[%d]:Error. The heap of problems is full.",$);
	 printf("\n[%d]: head = %d tail = %d",$,head,tail);
	 printf("\nIncrease the value of MAXHEAPSIZE. Current value is (%d)\n",MAXHEAPSIZE);
	 exit(1);
  }
  setProblem(&problems[tail+$],Prob.k, Prob.C, Prob.P);
  if ($ == 0) {
		tail = next;
	  farm printf("\n[%d]: inserting %d items, head = %d tail = %d",$,NUMPROCESSORS, head,tail);
	}
}


sync number numChildren(problemt Prob) {
  if (Prob.k < N) return 2; else return 0;
}

sync void setChild(problemt *child, problemt father,number order) {

  setProblem(child,
				 father.k+1,
				 father.C-(1-order)*w[father.k],
				 father.P+(1-order)*p[father.k]
				);
}

sync void branchAndBound(void) {
  problemt Prob, Probk;
  number nC, k, L, U;

  farm if ($ == 0) printf("\n starting b&b ...");
  while ((nP = numProblems()) > 0) {
    farm if ($ == 0) printf("\n inside while nP > 0 ...");
		NUMPROCESSORS = 0;
		$ = mpadd(&NUMPROCESSORS,1);
    farm if ($ == 0) printf("\n NUMPROCESSORS = %d",NUMPROCESSORS);
		if ($ < nP) {
		 getProblem(&Prob);
		 nC = numChildren(Prob);
		 if (nC != 0) { /* not a leaf */
			lowerUpper(Prob, &L, &U);
			if (bestSol < L) bestSol = L;	/* L <= bestSol <= U */
			if (bestSol < U) {
				for(k = 0; k < nC; k++) {
					setChild(&Probk, Prob, k);
					insertProblem(Probk);
				}
			}
		 } /* end nC != 0 */
		} /* end $ < nP */
		farm syncmax(&bestSol,bestSol);
  } /* end while */
}

int main(void) {
  problemt Prob;

	start {
		if ($ ==0) {
			readKnap();
			setProblem(&Prob, 0, M, 0);
			insertProblem(Prob);
		}
  	branchAndBound();
		seq {
			printf("\nsol = "); printf(ns, bestSol);
		}
  } /* end start */
}
